/****************************************************************************
** Demonstration of automatically detecting the Sound Blaster settings     **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************/

#include <dos.h>
#ifdef __cplusplus
  #include <conio.h>
  #include <stdio.h>
  #include <string.h>
#endif

  unsigned int          Base; //Sound Blaster base address
  char             CType[80]; //Card type
  unsigned char LoDMA, HiDMA; //DMA Channel and IRQ
  unsigned char          IRQ;
#ifdef __cplusplus
  void interrupt (*Old71)(...); //Pointers to old interrupt routines
  void interrupt (*Old11)(...);
  void interrupt (*Old13)(...);
  void interrupt (*Old15)(...);
  void interrupt (*Old72)(...);
  void interrupt (*Old73)(...);
#else
  void interrupt (*Old71)(); //Pointers to old interrupt routines
  void interrupt (*Old11)();
  void interrupt (*Old13)();
  void interrupt (*Old15)();
  void interrupt (*Old72)();
  void interrupt (*Old73)();
#endif

/****************************************************************************
** Checks to see if a Sound Blaster exists at a given address, returns     **
** true if Sound Blaster found, false if not.                              **
****************************************************************************/
char ResetDSP(unsigned int Test)
{
  //Reset the DSP
  outportb (Test + 0x6, 1);
  delay(10);
  outportb (Test + 0x6, 0);
  delay(10);
  //Check if reset was succesfull
  if (((inportb(Test + 0xE) & 0x80) == 0x80) && (inportb (Test + 0xA) == 0xAA)) {
    //DSP was found
    Base = Test;
    return (1);
  } else
    //No DSP was found
    return (0);
}

/****************************************************************************
** Read a byte from the DSP (Digital Signal Processor) on the Sound Blaster**
****************************************************************************/
unsigned char ReadDSP ()
{
  while (!(inportb (Base + 0xE) & 0x80));
  return (inportb (Base + 0xA));
}

/****************************************************************************
** Send a byte to the DSP (Digital Signal Processor) on the Sound Blaster  **
****************************************************************************/
void WriteDSP (unsigned char Value)
{
  //Wait for the DSP to be ready to accept data
  while ((inportb (Base + 0xC) & 0x80) == 0x80);
  //Send byte
  outportb (Base + 0xC, Value);
}

/****************************************************************************
** Returns the version number of the DSP.                                  **
****************************************************************************/
unsigned int DSPVersion ()
{
  unsigned int Version;
  WriteDSP (0xE1);
  Version = ReadDSP ();
  return (ReadDSP () + (Version << 8));
}

/****************************************************************************
** This tells the DSP to start playing a sample. However, the DMA          **
** controller has not been set up. Therefore the DSP signals the DMA       **
** controller for data. We check to see where that call takes place and    **
** thus determine the DMA channel the DSP uses                             **
****************************************************************************/
unsigned char TestLoDMA ()
{
  unsigned char Before, After;
  unsigned int          Count;

  //Reset DSP
  ResetDSP (Base);
  WriteDSP (0xD3);             //DSP-command D3h - Speaker Off
  WriteDSP (0x40);             //DSP-command 40h - Set sample frequency
  WriteDSP (165);              //Write time constant for 11025Hz
  WriteDSP (0x14);             //DSP-command 14h - 8bit playback
  WriteDSP (0x01);             //Set the block length to 1

  Count = 0;

  //Disable interrupts
  asm { cli }

  //Look at the current state of the DMA requests
  Before = inportb (0x8) & 0xE0;

  //Start playback, * note : this is still part of "set block length"
  WriteDSP (0x00);

  do {
    //Check them again until there is a change or it takes too long
    After = inportb (0x8) & 0xE0;
  } while ((Before == After) && (++Count < 60000));

  //Reenable interrupts
  asm { sti }

  //Reset DSP
  ResetDSP (Base);

  //Is there a change?
  if (Before != After) {
    Count = 0;
    After ^= (Before);
    //If yes, which bit is set?
    while (After > 0) { Count++; After >>= 1; }
    return (Count - 5);
  }
  //No DMA???
  return (0xFF);
}

/****************************************************************************
** This does the same thing for the 16bit DMA channel, if used             **
****************************************************************************/
unsigned char TestHiDMA ()
{
  unsigned char Before, After;
  unsigned int          Count;

  //Reset DSP
  ResetDSP (Base);
  WriteDSP (0xB0);             //DSP-command B0h - 16bit DMA
  WriteDSP (0x10);             //Mono
  WriteDSP (0x03);             //Set the block length to 3 = 1 DWORD

  Count = 0;

  //Disable interrupts
  asm { cli }

  //Look at the current state of the DMA requests
  Before = inportb (0xD0) & 0xF0;

  //Start playback, * note : this is still part of "set block length"
  WriteDSP (0x00);

  do {
    //Check them again until there is a change or it takes too long
    After = inportb (0xD0) & 0xF0;
  } while ((Before == After) && (++Count < 60000));

  //Reenable interrupts
  asm { sti }

  //Reset DSP
  ResetDSP (Base);

  //Is there a change?
  if (Before != After) {
    Count = 0;
    After ^= (Before);
    //If yes, which bit is set?
    while (After > 0) { Count++; After >>= 1; }
    return (Count - 1);
  }
  //No DMA???
  return (0xFF);
}

/****************************************************************************
** This tells the DSP to start playing a sample. And waits to see which    **
** one of the IRQ service routines is called.                              **
****************************************************************************/
void TestIRQ ()
{
  unsigned int Count;
  IRQ = 0;
  WriteDSP(0xD1);                    //DSP-command D1h - Enable speaker, required
                                     //on older SB's
  WriteDSP(0x40);                    //DSP-command 40h - Set sample frequency
  WriteDSP(165);                     //Write time constant for 11025Hz
  outportb (0x0A, 4 + LoDMA);        //Mask DMA channel
  outportb (0x0C, 0);                //Clear byte pointer
  outportb (0x0B, 0x49);             //Set mode
  outportb (1 + (LoDMA << 1), 0x03); //Set the block length to 3 == 1 DWORD
  outportb (1 + (LoDMA << 1), 0x00);

  outportb (0x0A, LoDMA);            //Unmask DMA channel

  WriteDSP(0x14);                    //DSP-command 14h - Single cycle playback
  WriteDSP(0x03);
  WriteDSP(0x00);

  Count = 0;
  do
    //Wait until either one of the IRQ service routines has been called
    //or it takes too long
    Count++;
  while ((!IRQ) && (Count != 0xFFFF));

}

/****************************************************************************
** IRQ service routine - this is called when the DSP has finished playing  **
** a block.                                                                **
****************************************************************************/
#ifdef __cplusplus
  void interrupt ServiceIRQ2 (...)
#else
  void interrupt ServiceIRQ2 ()
#endif
{
  //Relieve DSP
  inportb (Base + 0xE);
  //Acknowledge hardware interrupt
  outportb (0x20, 0x20);
  //Since IRQ 2 has been redirected, we need to acknowledge two interrupt
  //controllers
  outportb (0xA0, 0x20);
  //Save interrupt channel number
  IRQ = 2;
}

#ifdef __cplusplus
  void interrupt ServiceIRQ3 (...)
#else
  void interrupt ServiceIRQ3 ()
#endif
{
  inportb (Base + 0xE); outportb (0x20, 0x20); IRQ = 3;
}

#ifdef __cplusplus
  void interrupt ServiceIRQ5 (...)
#else
  void interrupt ServiceIRQ5 ()
#endif
{
  inportb (Base + 0xE); outportb (0x20, 0x20); IRQ = 5;
}

#ifdef __cplusplus
  void interrupt ServiceIRQ7 (...)
#else
  void interrupt ServiceIRQ7 ()
#endif
{
  inportb (Base + 0xE); outportb (0x20, 0x20); IRQ = 7;
}

#ifdef __cplusplus
  void interrupt ServiceIRQ10 (...)
#else
  void interrupt ServiceIRQ10 ()
#endif
{
  inportb (Base + 0xE); outportb (0x20, 0x20); outportb (0xA0, 0x20); IRQ = 10;
}

#ifdef __cplusplus
  void interrupt ServiceIRQ11 (...)
#else
  void interrupt ServiceIRQ11 ()
#endif
{
  inportb (Base + 0xE); outportb (0x20, 0x20); outportb (0xA0, 0x20); IRQ = 11;
}

void main ()
{
  int Temp;
  //Clear screen
  clrscr ();
  //Check for Sound Blaster, address: ports 220, 230, 240, 250, 260 or 280
  for (Temp = 1; Temp < 9; Temp++) {
    if (Temp != 7)
    if (ResetDSP (0x200 + (Temp << 4))) {
      break;
    }
  }
  if (Temp == 9) {
    //or none at all
    printf ("No Sound Blaster found\n");
    return;
  }

  //There are several different type of Sound Blaster
  if (DSPVersion () == 0x100) { strcpy (CType, "Sound Blaster"); }
  if (DSPVersion () == 0x105) { strcpy (CType, "Sound Blaster 1.5"); }
  if (DSPVersion () == 0x200) { strcpy (CType, "Sound Blaster Pro 2"); }
  if (DSPVersion () == 0x300) { strcpy (CType, "Sound Blaster Pro 3"); }
  if ((DSPVersion () >> 8) >= 4) { strcpy (CType, "Sound Blaster 16/ASP/AWE 32/AWE 64"); }

  //Check for DMA Channel
  LoDMA = TestLoDMA ();
  if ((DSPVersion () >> 8) >= 4) {
    HiDMA = TestHiDMA ();
    if (HiDMA == 0xFF) HiDMA = LoDMA;
  }

  //Save old IRQ vectors
  Old71 = getvect (0x71);
  Old11 = getvect (11);
  Old13 = getvect (13);
  Old15 = getvect (15);
  Old72 = getvect (0x72);
  Old73 = getvect (0x73);

  //Set new IRQ vectors
  setvect (0x71, ServiceIRQ2);
  setvect (11, ServiceIRQ2);
  setvect (13, ServiceIRQ5);
  setvect (15, ServiceIRQ7);
  setvect (0x72, ServiceIRQ10);
  setvect (0x73, ServiceIRQ11);

  //Enable IRQs
  outportb (0x21, inportb (0x21) & (!(1 << 2)));
  outportb (0x21, inportb (0x21) & (!(1 << 3)));
  outportb (0x21, inportb (0x21) & (!(1 << 5)));
  outportb (0x21, inportb (0x21) & (!(1 << 7)));
  outportb (0xA1, inportb (0xA1) & (!(1 << 1)));
  outportb (0xA1, inportb (0xA1) & (!(1 << 2)));
  outportb (0xA1, inportb (0xA1) & (!(1 << 3)));

  //Scan which IRQ is called
  TestIRQ ();

  //Free interrupt vectors used to service IRQs
  setvect (0x71, Old71);
  setvect (11, Old11);
  setvect (13, Old13);
  setvect (15, Old15);
  setvect (0x72, Old72);
  setvect (0x73, Old73);

  //Mask IRQs
  outportb (0x21, inportb (0x21) | (1 << 2));
  outportb (0x21, inportb (0x21) | (1 << 3));
  outportb (0x21, inportb (0x21) | (1 << 5));
  outportb (0x21, inportb (0x21) | (1 << 7));
  outportb (0xA1, inportb (0xA1) | (1 << 1));
  outportb (0xA1, inportb (0xA1) | (1 << 2));
  outportb (0xA1, inportb (0xA1) | (1 << 3));

  //Display results
  printf ("Card                : %s\n", CType);
  printf ("Base port           : %Xh\n", Base);
  printf ("IRQ                 : %d\n", IRQ);
  printf ("DMA channel (8bit)  : %d\n", LoDMA);
  if ((DSPVersion () >> 8) >= 4)
    printf ("DMA channel (16bit) : %d\n", HiDMA);
}